#include "Corr.h"
#include "assert.h"
#include <sstream>
#include <iostream>
#include <sys/stat.h>

Corr::Corr(System * sys, int NumWallGridPoints, double MiddleSep, double LayerSize) {
   this->sys = sys;
   this->NumWallGridPoints = NumWallGridPoints;
   double width = sys->getPolymers()->getGrid()->getBoxWidth().xcomp();
   InvMiddleSep = 1.0/MiddleSep;
   NumMiddleGridPoints = (int) (width*InvMiddleSep)+1;
   MakeHistGrids();
   this->LayerSize = LayerSize;
   this->InvLayerSep = NumWallGridPoints/LayerSize;
}

Corr::Corr(System * sys, int NumGTot) {
   this->sys = sys;
   int NumBtwn2s = 4;
   this->NumGTot = NumGTot;
#ifdef LOG_G_POINTS
   GArraySize = NumBtwn2s*(int)(log(NumGTot*1.0F)/log(2.0F));
#else
   GArraySize = NumGTot/2;
#endif
   G_t = new double[GArraySize];
   DeltaG_t = new double[GArraySize];
   GTimeArray = new int[GArraySize];
   NumGMeasurements = new int[GArraySize];
   for(int i=0;i < GArraySize; i++) {
#ifdef LOG_G_POINTS
      GTimeArray[i] = (int)(8.0*pow(2,((double) i)/NumBtwn2s));
#else
      GTimeArray[i] = 2*i;
#endif
   }
   GArrayIndex = 0;

   InitCoords = new vect_f[sys->getPolymers()->getTotMon()];

   NumGCalls = 0;
   NumCalls = 0;
   RgSq=0;
   REndSq=0;
   f_end_sum = 0;
   f_end_var = 0;
}

void Corr::setHeaderArgs(int const argc, const char ** argv) {
   this->argc = argc;
   this->argv = argv;
}

void Corr::setStreams(const char * density,const  char * stats) {
   ofDensity.open(density);
   ofStats.open(stats);
}

void Corr::setStreams(const char * dirname, const char * density,const  char * stats) {
   mkdir(dirname,S_IRWXU);
   ostringstream ostrdensity;
   ostrdensity << dirname << '/' << density;
   ofDensity.open(ostrdensity.str().c_str());
   ostringstream ostrstats;
   ostrstats << dirname << '/' << stats;
   cout << "output file: " << ostrstats.str().c_str() << endl;
   ofStats.open(ostrstats.str().c_str());
   assert(ofStats.good());
   ofStats << "writing to ofStats " << endl;
}


void Corr::MakeHistGrids() {
   HistLayer = new int[NumWallGridPoints];
   HistMiddle = new int[NumMiddleGridPoints];
}

void Corr::printHeader(ofstream & of)
{
   of << "# ";
   for(int i =0; i < argc ; i++) 
   {
      of << argv[i] << " ";
   }
   of << endl;
}


void Corr::addPointToHist(int * hist, double pos, double norm) {
   hist[(int)(pos*norm)]++;
}

void Corr::recordDensity(vect_f & r) {
   double x = r.xcomp();
   GridPeriodic * grid = sys->getPolymers()->getGrid();
   vect_f size = grid->getBoxWidth();
   double width = size.xcomp();

   double x_prime = width - x;

   if (x < LayerSize) 
      addPointToHist(HistLayer, x, InvLayerSep);
   if (x_prime < LayerSize) 
      addPointToHist(HistLayer, x_prime, InvLayerSep);

   addPointToHist(HistMiddle, x, InvMiddleSep);
   //addPointToHist(HistMiddle, x_prime, InvMiddleSep);

}

void Corr::updateDensity() {
   Polymers * polys = sys->getPolymers();
   Polymer * polymers = polys->getPolymers();
   for (int i=0; i < polys->getNumChains(); i++) {
      Polymer * poly = polymers + i;
      for(int j=0; j < poly->getNumMon(); j++) {
	 recordDensity(poly->getCoords()[j]);
      }
   }
}

void Corr::updateRg() {
   Polymers * polys = sys->getPolymers();
   RgSq += polys->RgSq();
   REndSq += sqrt(polys->REndEndSq());
}

void Corr::updateForce(){
   Polymers * polymers = sys->getPolymers();
   int NumChains = polymers->getNumChains();
   int NumMon = polymers->getNumMon();
   vect_f EndForce = polymers->getMommaCoords()[2*NumChains*NumMon]/(sys->getRK_StepSize()*(sys->getGigTime()));

   double fx = EndForce.xcomp();
   f_end_sum += fx;
   f_end_var += SQR(fx);

}


void Corr::update() {
   //can't be used as is with periodic a periodic system
//   updateDensity();
   updateRg();
   updateForce();
   if (NumGCalls == GTimeArray[GArrayIndex]) updateRiRiT();
   NumCalls++;
   NumGCalls++;
}


void Corr::updateRiRiT() {
   int TotMon = sys->getPolymers()->getTotMon();
   vect_f * coords_t = sys->getPolymers()->getMommaCoords();
   for (int i=0; i < TotMon; i++) {
      G_t[GArrayIndex] += SQR(coords_t[i] - InitCoords[i]);
   }
   DeltaG_t[GArrayIndex] += sqrt(SQR(coords_t[TotMon-1] -coords_t[0])*SQR(InitCoords[TotMon-1]-InitCoords[0]));
   NumGMeasurements[GArrayIndex]++;
   GArrayIndex++;
   if (GArrayIndex >= GArraySize) {
      GArrayIndex = 0;
      setInitCoords(sys->getPolymers()->getMommaCoords());
      NumGCalls=-1;
      sys->getPolymers()->adjustChain();
   }
}

void Corr::setInitCoords(vect_f * coords) {
   int TotMon = sys->getPolymers()->getTotMon();
   for (int i=0; i < TotMon; i++) {
      InitCoords[i] = coords[i];
   }
}


void Corr::printHists(ofstream & of) {
   double MiddleSep = 1.0/InvMiddleSep;
   double LayerPointSpacing = 1/InvLayerSep;

   printHeader(of);
   for(int i=0; i < NumWallGridPoints; i++) {
      of <<  LayerPointSpacing * (i+0.5) << " " <<  0.5*HistLayer[i]/(NumCalls*LayerPointSpacing) << endl;
   }
   of << "\n\n";
   for(int i=0; i < NumMiddleGridPoints-1; i++) {
      of <<  MiddleSep * (i+0.5) << " " <<  HistMiddle[i]/(NumCalls*MiddleSep) << endl;
   }
   of << "\n\n";
}

void Corr::printHists(){
   printHists(ofDensity);
}

void Corr::printStats(ofstream & of) {
   printHeader(of);
   int TotMon = sys->getPolymers()->getNumChains()*sys->getPolymers()->getNumMon();

   of << "# Rg^2 = " << RgSq/NumCalls << endl;
   double REnd2 = REndSq/NumCalls;
   of << "# REndEnd = " << REnd2  << endl;
   double f_ave = f_end_sum/NumCalls;
   f_end_var /= NumCalls;
   double stdeviation =  sqrt((f_end_var- SQR(f_ave))/NumCalls);
   cout << "# f_end_ave = " << f_ave << " std_dev_f = " << stdeviation << endl;
   of << "# f_end_ave = " << f_ave << " std_dev_f = " << stdeviation << endl;
   for(int i=0; i < GArraySize; i++) {
      of << GTimeArray[i] << " " << G_t[i]/(NumGMeasurements[i]*TotMon) << " " << 
	  (DeltaG_t[i]/NumGMeasurements[i]) - SQR(REnd2) << " " << NumGMeasurements[i] << endl;
   }
}

void Corr::printStatsCout() {
     printStats((ofstream &)cout);
}

void Corr::printStats() {
   assert(ofStats.good());
   printStats(ofStats);
}

void Corr::ofClose() {
   assert(ofDensity.good());
   ofDensity.close();
   ofStats.close();
}

