
from scipy import *
from scipy.linalg import *
from pylab import *
from scipy.integrate import *

#This code solves for the motion of a chain with 
#1. fixed link lengths, 
#2. inertia is #negligible (it always is for microscopic systems in a solvent) 
#3 A drag on each monomer proportional to the velocity
#4 Various other forces can be added to the simulation, like giving it an elastic
#  bending rigity, or pulling on the ends.
#5 It can be used to study electrophoresis by adding a periodic
# force field representing an array of obstacles.

#Because the links are constant in length, this makes it useful for studying many
#problems where the chain is close to being inextensible. However the algorithm
# to enforce inextensible links is a bit more involved. It involves calculating the
# tension along the chain as a function of position. Once the tension is calculated
# it is easy to add it to the other forces in the problem to get the equation of motion.

#This class is used to calculate the tension on the links of the chain.

class Tension:
    def __init__(self, T, n, d):
        self.T = T
        self.delta = zeros((n-1,d),dtype=double)
        self.delta_f = zeros((n-1,d),dtype=double)
        self.u = zeros((n-1),dtype=double)
        self.l = zeros((n-1),dtype=double)
        self.rhs = zeros((n-1,d),dtype=double)
        self.diag = zeros((n-1),dtype=double)
        self.force = zeros((n,d),dtype=double)


    def tension(self, r, f):
        self.delta = r[1:,:]-r[:-1,:]
        self.diag = -2*sum(self.delta*self.delta,axis=1)
        self.delta_f = f[1:,:]-f[:-1,:]
        u_vect = self.delta[1:,:]*self.delta[:-1,:]
        self.u[1:] = sum(u_vect, axis = 1)
        self.l[:-1] = self.u[1:]
        self.rhs = -sum(self.delta_f*self.delta,axis=1) - c1*(-0.5*self.diag - StepLengthSq)

        return solve_banded((1, 1), array((self.u,self.diag,self.l)), self.rhs, overwrite_ab=0, overwrite_b=0, debug=0)

    def calcForce(self,r,f):
        t = self.tension(r,f)
        for i in range(d):
            self.force[1:-1,i] = t[1:]*self.delta[1:,i] - t[0:-1]*self.delta[0:-1,i] + f[1:-1,i]
        self.force[0,:] = t[0]*self.delta[0,:] + f[0,:]
        self.force[n-1,:] = -t[n-2]*self.delta[n-2,:] + f[n-1,:]
        return self.force

def latticeForce(r):
    distsq[:] = 0
    for i in range(0,d-1):
        rem_r[:,i] = (remainder(a*10000.0+r[:,i],a)/a - 0.5)
        distsq[:] += rem_r[:,i]**2
    for i in range(0,d-1):
       f_lattice[:,i] = 0.01*signbit(distsq[:]-softcoresq) * rem_r[:,i] * (1.0/(distsq[:]-hardcoresq) - oneoverblah)**4
        
    if False and any(f_lattice != 0):
        print f_lattice
    return f_lattice


f_ext = array([7,5.0,0.0])
def dx_dt(r_in,t):
    r = reshape(r_in,(n,d))
    f = noise + latticeForce(r) + f_ext
    return tens.calcForce(r,f).flatten()

def length_error(r):
    delta = r[1:,:] - r[:-1,:]
    dlt_sq = delta*delta
    return sum(sum(dlt_sq)) - (n-1)

def construct_chain(n):
    r = zeros((1,d))
    r1 = zeros((1,d))
    shift = a*array([-10.0,-10.0,0.0])
    for i in range(n-1):
        dlt = array([0.0,1.0,0.0])

        r1 += reshape(dlt,(1,d))
        
        r = append(r, r1, 0)
    r += shift
    return r

   

#seed(3001)
seed(3002)
StepLength = 1.0
StepLengthSq = StepLength**2
#dimension of space
d = 3
#number of beads on the chain
n = 32
#time step used in diff eq solver:
dt = 0.05
#Temperature:
T = 1.0
c1 = 1 #helps to maintain constant link lengths

# Lattice spacing for the periodic force field.
a = 2.21
# lattice force field seen by every monomer
f_lattice = zeros((n,d))
distsq = zeros((n))
rem_r = zeros((n,d-1))

# Hard core radius for the lattice of obstacles
hardcore = 0.6/a
hardcoresq = hardcore**2
#  Outside the hard core is a region where the force decreases. It goes to zero at the "softcore" radius.
softcore = 1.0/a
softcoresq = softcore**2
oneoverblah = 1.0/(softcoresq-hardcoresq)


tens = Tension(T, n, d)

#r holds an array of the coordinates of the chain
r = construct_chain(n)

#time array
t = arange(0,2*dt,dt)

f = zeros((n,d))

graphics= True
if graphics:
    from visual import *
    ball = []
    cyl = []
    for i in range(n-1):
        cyl.append(cylinder (pos=r[i], axis=r[i+1]-r[i], color=color.red, radius = 0.2))

    for x in range(-10,25):
        for y in range(-10,30):
            cylinder(pos=(a*(x+0.5),a*(y+0.5),0), axis =(0,0,5.0), color=color.green, radius=hardcore)

if graphics:
    ion()


x_out = r.flatten()

x_out = reshape(x_out,(1,n*d))

display_interval = 2
time = 0;
for t_coarse in arange(0,100,dt):
    if graphics:
        rate(10000)
    t = array([t_coarse,t_coarse+dt])
    #thermal noise is generated and used to push the chain:
    noise = sqrt(2*T/dt)*standard_normal((n,d))
    #The diff'lequations are integrated. All the work comes in calculating dx_dt
    x_out = odeint(dx_dt, x_out[-1] , t, full_output=False)
    time += 1
    if time%300 == 0 : 
        #check to make sure the links haven't changed their lengths
        print length_error(reshape(x_out[-1],(n,d)))
    if  time%display_interval == 0 and graphics:
        r_out = reshape(x_out[-1],(n,d))
        for i in range(n-1):
            cyl[i].pos = r_out[i]
            cyl[i].axis = r_out[i+1]-r_out[i]



