
from scipy import *
from pylab import *
from scipy.integrate import *

# Simulation of Hodgkin and  Huxley's (1952) model (HH) of membrane currents in a nerve

#  n,m,h are "gating" variables with:
#  n for potassium (K)
#  m for sodium (Na)
#  h for other ions

Cm=1
#conductance constants, in HH these have bars on the top:
g_K =  36.0
g_Na = 120.0
g_l =  0.3
# Voltage Displacements 
V_Na = -115.0
VK = 12.0
Vl = -10.613

#Can turn the conductances off for testing:
#g_K=0
#g_Na = 0
#g_l = 0

#Note units for I: g is in units milli/(ohm cm&2), V is in mV, so I has units 10^{-5} Amps/cm^2
#Vt = dV/dt, V is the potential:
def I(Vt, V, n, m, h):
    return Cm*Vt + g_K*(n**4)*(V-VK)+ g_Na*(m**3)*h*(V-V_Na)+ g_l*(V-Vl)



def alpha_n(V, a_0 = 0.01, V_n = 10.0, V_shift = 10.0):
    if V != -V_shift:
         return a_0*(V+V_shift)/(exp((V+V_shift)/V_n)-1.0)
    else:
         return a_0*V_n 

def beta_n(V, b_0 = 0.125, V_N = 80.0):
    return b_0*exp(V/V_N)


def alpha_m(V):
    return  alpha_n(V, a_0 = 0.1, V_n = 10.0, V_shift = 25.0)

def beta_m(V):
    return beta_n(V,b_0 = 4.0, V_N = 18.0)

def alpha_h(V):
    return beta_n(V, b_0=0.07, V_N=20.0)

V_shift_h = 30.0
v_n = 10.0
def beta_h(V):
    return 1.0/(exp((V+V_shift_h)/v_n)+1.0)


def dx_dt(x_in,t):
    n = x_in[0] 
    dxdt[0] = alpha_n(V)*(1-n)-beta_n(V)*n
    m = x_in[1]
    dxdt[1] = alpha_m(V)*(1-m)-beta_m(V)*m
    h = x_in[2]
    dxdt[2] = alpha_h(V)*(1-h)-beta_h(V)*h
    return dxdt

dxdt = zeros((3))
total_time = 30.0
dt = 0.01
t = arange(0,total_time,dt)

V_array = arange(-110.,10.)

def arr_func(func,xrnge):
    count = 0
    arr = zeros(shape(xrnge))
    for x in xrnge:
        arr[count] = func(xrnge[count])
        count += 1
    return arr

def plot_alphas():
    figure(0)
    plot(V_array,arr_func(alpha_n,V_array),label="alpha_n")
    plot(V_array,arr_func(beta_n,V_array),label="beta_n")
    legend()
    title('n: on/off rates ')
    draw()

    figure(1)
    plot(V_array,arr_func(alpha_m,V_array),label="alpha_m")
    plot(V_array,arr_func(beta_m,V_array),label="beta_m")
    legend()
    title('m: on/off rates ')
    draw()

    figure(2)
    plot(V_array,arr_func(alpha_h,V_array),label="alpha_h")
    plot(V_array,arr_func(beta_h,V_array),label="beta_h")
    xlabel('V')
    legend()
    title('h: on/off rates ')
    draw()

plot_alphas()
V0 = 0

n0 = alpha_n(V0)/(alpha_n(V0)+beta_n(V0))
m0 = alpha_m(V0)/(alpha_m(V0)+beta_m(V0))
h0 = alpha_h(V0)/(alpha_h(V0)+beta_h(V0))

x_init = array([n0,m0,h0])




V=-25.0


x_out = odeint(dx_dt, x_init , t, full_output=0)

figure(3)
plot(t,x_out[:,0],label='n')
plot(t,x_out[:,1],label='m')
plot(t,x_out[:,2],label='h')
xlabel('t')
legend()
title('Gating Variables')
draw()

figure(4)
plot(t,g_K*x_out[:,0]**4,label='K')
plot(t,g_Na*x_out[:,1]**3*x_out[:,2],label='Na')
plot(t,g_l*ones(size(t)),label='Oth')
xlabel('t')
legend()
title('Conductances')
draw()


figure(5)
i = zeros((size(t)))
count = 0
for time in t:
    i[count] = I(0,V,x_out[count,0],x_out[count,1],x_out[count,2])
    count += 1
plot(t,i,label='I')
xlabel('t')
legend()
title('Current')
draw()

show()


